/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 * 
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 * 
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License. You can obtain
 * a copy of the License at https://glassfish.dev.java.net/public/CDDL+GPL.html
 * or glassfish/bootstrap/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 * 
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at glassfish/bootstrap/legal/LICENSE.txt.
 * Sun designates this particular file as subject to the "Classpath" exception
 * as provided by Sun in the GPL Version 2 section of the License file that
 * accompanied this code.  If applicable, add the following below the License
 * Header, with the fields enclosed by brackets [] replaced by your own
 * identifying information: "Portions Copyrighted [year]
 * [name of copyright owner]"
 * 
 * Contributor(s):
 * 
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */
package org.glassfish.admin.amx.core;

import org.glassfish.admin.amx.util.jmx.JMXUtil;
import org.glassfish.admin.amx.util.SetUtil;
import org.glassfish.admin.amx.util.TypeCast;
import org.glassfish.admin.amx.util.ClassUtil;
import org.glassfish.api.amx.AMXMBeanMetadata;

import static org.glassfish.admin.amx.core.AMXConstants.*;

import javax.management.Notification;
import javax.management.ObjectName;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import javax.management.MBeanServer;

/**
	Utility routines pertinent to the MBean API.
 */
public final class Util
{
    private static void debug( final String s ) { System.out.println(s); }
    
	private	Util()	{}
	
	/**
		Create a new ObjectName, caller is guaranteeing that the name is
		well-formed (a RuntimeException will be thrown if not). This avoids
		having to catch all sorts of JMX exceptions.<br>
		<b>NOTE:</b> Do not call this method if there is not certainty of a well-formed name.
		
		@param name
	 */
		public static ObjectName
	newObjectName( String name )
	{
		return( JMXUtil.newObjectName( name ) );
	}
	
	
	/**
		Build an ObjectName.  Calls newObjectName( domain + ":" + props )
		
		@param domain	the JMX domain
		@param props	properties of the ObjectName
	 */
		public static ObjectName
	newObjectName( String domain, String props )
	{
		return( newObjectName( domain + ":" + props ) );
	}
	
	/**
		Build an ObjectName pattern.  
		
		@param domain	the JMX domain
		@param props	properties of the ObjectName
	 */
		public static ObjectName
	newObjectNamePattern(
		final String domain,
		final String props )
	{
		return( JMXUtil.newObjectNamePattern( domain, props ) );
	}
	
	/**
		Build an ObjectName pattern.
		
		@param objectName
	 */
		public static ObjectName
	newObjectNamePattern( ObjectName objectName )
	{
		final String	props	= objectName.getKeyPropertyListString();
		
		return( newObjectNamePattern( objectName.getDomain(), props) );
	}
	
	
	/**
		Make an ObjectName property of the form <i>name</i>=<i>value</i>.
		
		@param name
		@param value
	 */
		public static String
	makeProp(
		final String name,
		final String value )
	{
		return( JMXUtil.makeProp( name, value ) );
	}
	
	/**
		Make an ObjectName property of the form type=<i>value</i>.
		
		@param value
	 */
		public static String
	makeTypeProp( final String value )
	{
		return( makeProp( TYPE_KEY, value ) );
	}
	
	/**
		Make an ObjectName property of the form name=<i>value</i>.
	 */
		public static String
	makeNameProp( final String name )
    {
		return( makeProp( NAME_KEY, "" + name) );
	}

	/**
		@param type
		@param name
	 */
		public static String
	makeRequiredProps(
		final String type,
		final String name )
	{
        String props = Util.makeTypeProp( type );
        if ( ! (name == null || name.length() == 0 || name.equals(NO_NAME)) )
        {
            final String nameProp	= Util.makeNameProp( name );
            props	= Util.concatenateProps( props, nameProp );
        }
		
		return( props );
	}
	
	/**
		Extract the type and name properties and return it as a single property
		<i>type</i>=<i>name</i>
		
		@param objectName
	 */
		public static String
	getSelfProp( final ObjectName objectName )
	{
		final String	type	= objectName.getKeyProperty( TYPE_KEY );
		final String	name		= objectName.getKeyProperty( NAME_KEY );
		
		return( Util.makeProp( type, name ) );
	}
    
    /**
		Extract all properties other than type=<type>,name=<name>.
		
		@param objectName
	 */
		public static String
	getAdditionalProps( final ObjectName objectName )
	{
        final java.util.Hashtable allProps = objectName.getKeyPropertyList();
        allProps.remove( TYPE_KEY );
        allProps.remove( NAME_KEY );
        
        String props = "";
        for( final Object key : allProps.keySet() )
        {
            final String prop = makeProp( (String)key, (String)allProps.get(key) );
            props = concatenateProps( props, prop );
        }
		
		return props;
	}
	
	
	/**
		Get properties corresponding to the FullType of this ObjectName.
		Its type/name are included as <i>type</i>=<i>name</i>, <b>not</b>
		as type=<i>type</i>,name=<i>name</i>.
		
		@param objectName
		@param fullType
		@return String of relevant ObjectName properties
		public static String
	getFullTypeProps(
		final ObjectName	objectName,
		final String		fullType )
	{
		final String selfProp	= Util.getSelfProp( objectName );
			
		// now add properties for ancestors; skip the last type; it's
		// present in the type/name properties (selfProp)
		String	ancestorProps	= "";
		final String[]	types	= Util.getTypeArray( fullType );
		for( int i = 0; i < types.length - 1; ++i )
		{
			final String	key		= types[ i ];
			final String	value	= objectName.getKeyProperty( key );
			final String	prop	= Util.makeProp( key, value );
			
			ancestorProps	= Util.concatenateProps( ancestorProps, prop );
		}
		
		final String	props	= Util.concatenateProps( selfProp, ancestorProps );
		
		return( props );
	}
	 */
	
    
    /*
		public static String
	getJ2EEType( final Class<? extends AMXProxy> amxInterface )
	{
        final String fieldName = "J2EE_TYPE";
        
        try {
            final java.lang.reflect.Field field = amxInterface.getField( fieldName );
            return String.class.cast( field.get(null) );
        }
        catch( NoSuchFieldException e )
        {
            throw new RuntimeException( "Missing J2EE_TYPE field in interface " + amxInterface.getNameProp(), e);
        }
        catch( IllegalAccessException e )
        {
            throw new RuntimeException( "Can't access J2EE_TYPE field in " + amxInterface.getNameProp(), e);
        }
	}
    */
		
	/**
		Minimal ObjectName properties required for an ObjectName pattern
		to uniquely identify an MBean.  See {@link #getObjectNamePattern}.
	 */
	private static final Set<String> PATTERN_PROPS	=
	    SetUtil.newUnmodifiableStringSet(
		TYPE_KEY,
		NAME_KEY );


	/**
		Get all keys required for an ObjectName pattern which uniquely
		identifies the MBean.
		public static Set<String>
	getPatternKeys(
		final String		fullType )
	{
		final Set<String> requiredKeys	= SetUtil.copySet( PATTERN_PROPS );
		
		// omit the last one, it is the simple type of this MBean, which we've
		// already included
		final String[]	types	= Util.getTypeArray( fullType );
		for( int i = 0; i < types.length - 1; ++i )
		{
			requiredKeys.add( types[ i ] );
		}

		return TypeCast.checkedStringSet( requiredKeys  );
	}
	 */
	
		public static String
	concatenateProps(
		final String props1,
		final String props2 )
	{
		return( JMXUtil.concatenateProps( props1, props2 ) );
	}
	
		public static String
	concatenateProps(
		final String props1,
		final String props2,
		final String props3 )
	{
		return( concatenateProps( concatenateProps( props1, props2), props3) );
	}
	
	/**
		@return a Set of ObjectNames from a Set of AMX.
	 */
		public static List<ObjectName>
	toObjectNames( final Collection<? extends AMXProxy> amxs )
	{
		final List<ObjectName>	objectNames	= new ArrayList<ObjectName>();
		for( final AMXProxy next : amxs )
		{
			objectNames.add( getObjectName( next ) );
		}
		return( Collections.checkedList(objectNames, ObjectName.class) );
	}
	
	/**
		@return a Map of ObjectNames from a Map whose values are AMX.
	 */
		public static Map<String,ObjectName>
	toObjectNames( final Map<String,? extends AMXProxy> amxMap )
	{
		final Map<String,ObjectName>	m	= new HashMap<String,ObjectName>();
		
		for( final String key : amxMap.keySet() )
		{
			final AMXProxy	value	= amxMap.get( key );
			m.put( key, getObjectName(value) );
		}
		return( Collections.checkedMap( m, String.class, ObjectName.class) );
	}
	
	/**
		@return an ObjectName[] from an AMX[]
	 */
		public static ObjectName[]
	toObjectNames( final AMXProxy[] amx )
	{
		final ObjectName[]	objectNames	= new ObjectName[ amx.length ];
		for( int i = 0; i < objectNames.length; ++i )
		{
			objectNames[ i ]	= amx[ i ] == null ? null : getObjectName( amx[ i ] );
		}
		
		return( objectNames );
	}
	
	/**
		Extract the names from all ObjectNames.  The name is the value of the
		property NAME_KEY (See {@link AMXProxy}).  Note that if two or more ObjectNames
		share the same name, the resulting Set will be of smaller size() than
		the original.
		
		@return Set
		public static Set<String>
	getNames( final Set<? extends AMXProxy>	amxs )
	{
		return getNamesSet( Util.toObjectNames( amxs ) );
	}
	 */
	
	/**
		Extract the names from all ObjectNames.  The name is the value of the
		property NAME_KEY (See {@link AMXProxy}).  Note that if two or more ObjectNames
		share the same name, the resulting Set will be of smaller size() than
		the original.
		
		@return Set
		public static Set<String>
	getNamesSet( final Set<ObjectName>	objectNames )
	{
		return TypeCast.checkedStringSet(
		        JMXUtil.getKeyPropertySet( NAME_KEY, objectNames ) );
	}
	 */
	
	/**
		Extract the names from all ObjectNames.
		
		@return String[] of names from the ObjectNames
		public static String[]
	getNamesArray( final Set<ObjectName>	objectNames )
	{
		return( JMXUtil.getKeyProperty( NAME_KEY, objectNames ) );
	}
	 */
	
	/**
		Create a Map keyed by the value of the NAME_KEY with
		value the ObjectName. Note that if two or more ObjectNames
		share the same name, the resulting Map will contain only
		one of the original ObjectNames.
		
		@param objectNames Set of ObjectName
	 */
		public static final Map<String,ObjectName>
	createObjectNameMap( final Set<ObjectName> objectNames )
	{
		final Map<String,ObjectName>	m	= new HashMap<String,ObjectName>();
		
		for( final ObjectName objectName : objectNames )
		{
			final String		name	= getNameProp( objectName );
			
			assert( ! m.containsKey( name ) ) :
				"createObjectNameMap: key already present: " + name + " in " + objectName;
			m.put( name, objectName );
		}
		
		assert( m.keySet().size() == objectNames.size() );
		
		return( Collections.checkedMap(m, String.class, ObjectName.class) );
	}
	
	
	/**
		Create a Map keyed by the value of the NAME_KEY with
		value the AMX item.
		
		@param amxs Set of AMX
	 */
		public static <T extends AMXProxy> Map<String,T>
	createNameMap( final Set<T> amxs )
	{
		final Map<String,T>	m	= new HashMap<String,T>();
		
		for( final T amx : amxs )
		{
			final String	name	= amx.getName();
			m.put( name, amx );
		}
		
		return( m );
	}
    
	
	/**
		Get the ObjectName targeted by this {@link AMXProxy}.
	 */
		public static <T extends AMXProxy> ObjectName
	getObjectName( final T amx )
	{
		return  amx.extra().objectName();
	}
	
	
	/**
		All Notifications emitted by AMX MBeans which are not
		standard types defined by JMX place a Map
		into the userData field of the Notification.  This call
		retrieves that Map, which may be null if no additional
		data is included.
	 */
		public static Map<String,Serializable>
	getAMXNotificationData( final Notification notif )
	{
	    return Collections.unmodifiableMap(
	            JMXUtil.getUserDataMapString_Serializable( notif ) );
	}
	
	/**
	    Use of generic type form taking Class<T> is preferred.
	 */
	    public static Serializable
	getAMXNotificationValue( final Notification notif, final String key )
	{
		final Map<String,Serializable> data	=
		    getAMXNotificationData( notif );
		
		if ( data == null )
		{
			throw new IllegalArgumentException( notif.toString() );
		}
		
		if ( ! data.containsKey( key ) )
		{
			throw new IllegalArgumentException( "Value not found for " + key +
				" in " + notif );
		}
		
		return data.get( key );
	}
	
	/**
		Retrieve a particular value associated with the specified
		key from an AMX Notification.
		@see #getAMXNotificationData
	 */
		public static <T extends Serializable> T
	getAMXNotificationValue(
	    final Notification  notif,
	    final String        key,
	    final Class<T>      theClass)
	{
	    final Serializable value    = getAMXNotificationValue( notif, key );
		
		return theClass.cast( value );
	}
	
	    public static void
	sleep( final long millis )
	{
		try
		{
			Thread.sleep( millis );
		}
		catch( InterruptedException e )
		{
		}
	}
	
	
    /**
        A safe way to cast to AMX.
     */
        public static AMXProxy
    asAMX( final Object o)
    {
        return AMXProxy.class.cast( o );
    }
    
    /**
        Filter the AMX dynamic proxies to those that implement the specified interface,
        and return a new Set with the matching items.  The 'desired' interface can be
        any AMX-defined class, including the mixin ones.
         
        @param candidates the Set to be filtered
        @param desired the interface to filter by
     */
        public static <T extends AMXProxy> Set<T>
    filterAMX( final Set<T> candidates, final Class<?> desired )
    {
        final Set<T>   result = new HashSet<T>();
        for( final T amx : candidates )
        {
            if ( desired.isAssignableFrom( amx.getClass() ) )
            {
                result.add(amx);
            }
        }
        return result;
    }

        public static Map<String,ObjectName>
    filterByType(final ObjectName[] objectNames, final String type)
    {
        Map<String,ObjectName> m = null;
        if ( objectNames != null)
        {
            m = new HashMap<String,ObjectName>();
            for ( final ObjectName o : objectNames)
            {
                if ( type.equals( o.getKeyProperty(TYPE_KEY) ))
                {
                    // use the name property, not the name from getNameProp()
                    m.put(o.getKeyProperty(NAME_KEY), o);
                }
            }
        }
        return m;
    }
    
    
    /**
        Filter the AMX dynamic proxies to those that implement the specified interface,
        and return a new Map with the matching items.  The 'desired' interface can be
        any AMX-defined class, including the mixin ones.
         
        @param candidates the Map to be filtered
        @param desired the interface to filter by
     */
        public static <T extends AMXProxy> Map<String, T>
    filterAMX( final Map<String,T> candidates, final Class<?> desired )
    {
        final Map<String,T>   result = new HashMap<String,T>();
        for( final String key : candidates.keySet() )
        {
            final T amx = candidates.get(key);
            if ( desired.isAssignableFrom( amx.getClass() ) )
            {
                result.put(key, amx);
            }
        }
        return result;
    }
    
    
         public static String
    getTypeProp( final ObjectName objectName )
    {
        return objectName.getKeyProperty(TYPE_KEY);
    }
    
	/**
		Get the value of the NAME_KEY property within the ObjectName, or null if
        not present.
		@return the name
	 */
		public static String
	getNameProp( final ObjectName	objectName )
	{
		return objectName.getKeyProperty( NAME_KEY );
	}
    
		public static String
	getParentPathProp( final ObjectName objectName )
	{
		return objectName.getKeyProperty( PARENT_PATH_KEY );
	}
    
		public static String
	getParentPathProp( final AMXProxy amx )
	{
		return getParentPathProp( amx.extra().objectName() );
	}
    
        public static ObjectName
    getParent(final MBeanServer server, final ObjectName objectName )
    {
        return (ObjectName)JMXUtil.getAttribute( server, objectName, AMXConstants.ATTR_PARENT );
    }
    
    /**
        Deduce the type to be used in the path.  Presence of a type field always
        take precedence, then the AMXMBeanMetadata.
     */
        public static String
    deduceType( final Class<?> intf )
    {
        String type = null;
        
        AMXMBeanMetadata meta = null;
        final Object typeField = ClassUtil.getFieldValue( intf, TYPE_FIELD );
        if ( typeField instanceof String )
        {
            type = (String)typeField;
        }
        else if ( (meta = intf.getAnnotation(AMXMBeanMetadata.class)) != null )
        {
            final String typeValue = meta.type();
            
            if ( typeValue.equals( AMXMBeanMetadata.NULL ) || typeValue.length() == 0)
            {
                type = ClassUtil.stripPackageName( intf.getName() );
            }
            else
            {
                type = typeValue;
            }
        }
        else
        {
            // no annotation, use our default conversion
            type = ClassUtil.stripPackageName( intf.getName() );
        }
        return type;
    }
    
        public static ObjectName
    getAncestorByType( final MBeanServer mbeanServer, final ObjectName child, final String type )
    {
        ObjectName cur = child;
        
         while ( (cur = (ObjectName)JMXUtil.getAttribute( mbeanServer, cur, AMXConstants.ATTR_PARENT )) != null )
         {
            if ( getTypeProp(cur).equals(type) )
            {
                break;
            }
         }
         return cur;
    }
}























